#ifndef _EDITWING_VIEW_H_
#define _EDITWING_VIEW_H_
#include "ewCommon.h"
#include "ewDoc.h"
#ifndef __ccdoc__
namespace editwing {
namespace view {
#endif



class Canvas;
class ViewImpl;
class Cursor;
class Caret;



//=========================================================================
//@{ @pkg editwing.View //@}
//@{
//	`揈Ȃ
//
//	̃NXł́AbZ[W̕zsŁA
//	Canvas/ViewImpl ōsB̂ŁAڂ͂QƂ̂ƁB
//@}
//=========================================================================

class View : public ki::WndImpl, public doc::DocEvHandler
{
public:

	//@{ ȂRXgN^ //@}
	View( doc::Document& d, HWND wnd );
	~View();

	//@{ ܂Ԃؑ //@}
	void SetWrapType( int wt );

	//@{ sԍ\/\ؑ //@}
	void ShowLineNo( bool show );

	//@{ \FEtHgؑ //@}
	void SetFont( const VConfig& vc );

	//@{  //@}
	ViewImpl& impl() { return *impl_; }

	//@{ J[\ //@}
	Cursor& cur();

private:

	doc::DocImpl&      doc_;
	ki::dptr<ViewImpl> impl_;
	static ClsName     className_;

private:

	void    on_create( CREATESTRUCT* cs );
	void    on_destroy();
	LRESULT on_message( UINT msg, WPARAM wp, LPARAM lp );
	void    on_text_update( const DPos& s, const DPos& e, const DPos& e2, bool bAft, bool mCur );
	void    on_keyword_change();
};



//=========================================================================
//@{
//	CxgnhC^[tFCX
//
//	J[\甭CxgFX
//@}
//=========================================================================

class CurEvHandler
{
	friend class Cursor;
	virtual void on_move( const DPos& c, const DPos& s ) {}
	virtual void on_char( Cursor& cur, unicode wch );
	virtual void on_key( Cursor& cur, int vk, bool sft, bool ctl );
	virtual void on_ime( Cursor& cur, unicode* str, ulong len );
};



//=========================================================================
//@{
//	\ʒu܂Ŋ܂߂DPos
//@}
//=========================================================================

struct VPos : public DPos
{
	ulong vl; // VLine-Index
	ulong rl; // RLine-Index
	int   vx; // XN[lȂzXN[xW(pixel) 
	int   rx; // ̕тɍEĂȂxW(pixel)
		      //   == ŝۂZs [] ňړ
		      //   == ̌ [] Ŗ߂悤ȃAłB
	void operator=( const DPos& dp ) { tl=dp.tl, ad=dp.ad; }

	VPos(bool) : DPos(0,0),vl(0),rl(0),vx(0),rx(0) {}
	VPos() {}
};



//=========================================================================
//@{
//	J[\
//@}
//=========================================================================

class Cursor : public ki::Object
{
public:

	// Ƃ
	Cursor( HWND wnd, ViewImpl& vw, doc::DocImpl& dc );
	~Cursor();
	void AddHandler( CurEvHandler* ev );
	void DelHandler( CurEvHandler* ev );

	// J[\ړ
	void MoveCur( const DPos& dp, bool select );

	// L[ɂJ[\ړ
	void Left( bool wide, bool select );
	void Right( bool wide, bool select );
	void Up( bool wide, bool select );
	void Down( bool wide, bool select );
	void Home( bool wide, bool select );
	void End( bool wide, bool select );
	void PageUp( bool select );
	void PageDown( bool select );

	// eLXg
	void Input( const unicode* str, ulong len );
	void Input( const char* str, ulong len );
	void InputChar( unicode ch );
	void Del();
	void DelBack();

	// Nbv{[h
	void Cut();
	void Copy();
	void Paste();

	// [hؑ
	void SetInsMode( bool bIns );
	void SetROMode( bool bRO );

public:

	bool isInsMode() const;
	bool isROMode() const;
	bool isSelected() const;
	bool getCurPos( const VPos** start, const VPos** end ) const;
	void ResetPos();
	void on_scroll_begin();
	void on_scroll_end();
	void on_text_update( const DPos& s, const DPos& e, const DPos& e2, bool mCur );
	void on_setfocus();
	void on_killfocus();
	void on_keydown( int vk, LPARAM flag );
	void on_char( TCHAR ch );
	void on_ime_composition( LPARAM lp );
	void on_lbutton_down( short x, short y, bool shift );
	void on_mouse_move( short x, short y );
	void on_lbutton_up( short x, short y );
	void on_lbutton_dbl( short x, short y );
	bool on_contextmenu( short x, short y );
	void on_timer();

private:

	doc::DocImpl&   doc_;
	ViewImpl&       view_;
	CurEvHandler*   pEvHan_;
	ki::dptr<Caret> caret_;

	VPos cur_;  // J[\ʒu
	VPos sel_;  // I̎ʒu
	bool bIns_; // }[hH
	bool bRO_;  // ǎpH

	UINT timerID_;    // }EXhbOp
	int  keyRepTime_; // ^C}[֌W
	int  dragX_;      // ʒu
	int  dragY_;      // ʒu
	bool lineSelectMode_; // sI[hH

	CurEvHandler defaultHandler_;

private:

	void MoveByMouse( int x, int y );
	void MoveTo( const VPos& vp, bool sel );
	void Ud( int dy, bool select );
	void UpdateCaretPos();
	void Redraw( const VPos& s, const VPos& e );
};



//-------------------------------------------------------------------------

inline bool Cursor::isSelected() const
	{ return cur_!=sel_; }

inline bool Cursor::isInsMode() const
	{ return bIns_; }

inline bool Cursor::isROMode() const
	{ return bRO_; }



//=========================================================================

}}    // namespace editwing::view
#endif // _EDITWING_VIEW_H_
