#ifndef _KILIB_KTL_ARRAY_H_
#define _KILIB_KTL_ARRAY_H_
#include "types.h"
#include "memory.h"
#ifndef __ccdoc__
namespace ki {
#endif



//=========================================================================
//@{ @pkg ki.KTL //@}
//@{
//	{^p̉ϒz
//
//	ɊȑfȍłBANZX͍DȂƂ֎Rɉ\łA
//	vf̒ǉE폜͖ɑ΂Ă̂݁Brbg̃Rs[ƂA
//	FXGLZgbNȍƂĂ܂̂ŁAƂėǂ
//	^ȊOɂ͎gȂłB
//@}
//=========================================================================

template <typename T>
class storage : public Object
{
public:

	//@{
	//	RXgN^
	//
	//	@param alloc_size
	//		ŏɊmۂ""TCYB
	//		"z"TCYł͂ȂƂɒӁB
	//@}
	explicit storage( ulong allocSize=20 )
		: alen_( Max( allocSize, 1UL ) )
		, len_ ( 0 )
		, buf_ ( static_cast<T*>(mem().Alloc(alen_*sizeof(T))) )
		{}

	~storage()
		{ mem().DeAlloc( buf_, alen_*sizeof(T) ); }

	//@{ ɗvfǉ //@}
	void Add( const T& obj )
		{
			if( len_ >= alen_ )
				ReAllocate( alen_<<2 );
			buf_[ len_++ ] = obj;
		}

	//@{
	//	zTCYύX
	//
	//	k/ĝǂ\BRXgN^ƈႢA
	//	w肵lɊÂőindexω܂B
	//	@param new_size VTCYB
	//@}
	void ForceSize( ulong newSize )
		{
			if( newSize > alen_ )
				ReAllocate( newSize );
			len_ = newSize;
		}

public:

	//@{ vf	//@}
	ulong size() const
		{ return len_; }

	//@{ vf擾 //@}
	T& operator[]( ulong i )
		{ return buf_[i]; }

	//@{ vf擾(const) //@}
	const T& operator[]( ulong i ) const
		{ return buf_[i]; }

	//@{ z擪̃|C^Ԃ //@}
	const T* head() const
		{ return buf_; }

private:

	void ReAllocate( ulong siz )
		{
			ulong p = alen_*sizeof(T);
			T* newbuf = static_cast<T*>
				(mem().Alloc( (alen_=siz)*sizeof(T) ));
			memcpy( newbuf, buf_, len_*sizeof(T) );
			mem().DeAlloc( buf_, p );
			buf_ = newbuf;
		}

private:

	ulong alen_;
	ulong len_;
	T*    buf_;

private:

	NOCOPY(storage<T>);
};



//=========================================================================
//@{
//	IuWFNg^ɂgPXg
//
//	قƂǉo܂Bô͖ւ̒ǉƁAƎ
//	iteratorɂV[PVȃANZX̂݁B
//@}
//=========================================================================

template <class T>
class olist : public Object
{
private:

	struct Node {
		Node( const T& obj )
			: obj_ ( obj ), next_( NULL ) {}
		~Node()
			{ delete next_; }
		Node* Add( Node* pN )
			{ return next_==NULL ? next_=pN : next_->Add(pN); }
		T     obj_;
		Node* next_;
	};

public:

	struct iterator {
		iterator( Node* p=NULL ) : ptr_(p)   {}
		T& operator*()                       { return ptr_->obj_; }
		T* operator->() const                { return &ptr_->obj_; }
		bool operator==( const iterator& i ) { return i.ptr_==ptr_; }
		bool operator!=( const iterator& i ) { return i.ptr_!=ptr_; }
		iterator& operator++()    { ptr_=ptr_->next_; return *this; }
	private:
		Node* ptr_;
	};

public:

	//@{ RXgN^ //@}
	olist()
		: top_( NULL ) {}

	//@{ fXgN^ //@}
	~olist()
		{ empty(); }

	//@{ ɂ //@}
	void empty()
		{ delete top_; top_ = NULL; }

	//@{ 擪 //@}
	iterator begin()
		{ return iterator(top_); }

	//@{  //@}
	iterator end()
		{ return iterator(); }

	//@{ ɗvfǉ //@}
	void Add( const T& obj )
		{
			Node* pN = new Node( obj );
			(top_ == NULL) ? top_=pN : top_->Add( pN );
		}

	//@{ wvf폜 //@}
	void Del( iterator d )
		{
			if( d != end() )
			{
				Node *p=top_, *q=NULL;
				for( ; p!=NULL; q=p,p=p->next_ )
					if( &p->obj_ == &*d )
						break;
				if( q != NULL )
					q->next_ = p->next_;
				else
					top_ = p->next_;
				p->next_ = NULL;
				delete p;
			}
		}

	//@{ wvfȍ~SĂ폜 //@}
	void DelAfter( iterator d )
		{
			if( d != end() )
				if( d == begin() )
				{
					empty();
				}
				else
				{
					Node *p=top_, *q=NULL;
					for( ; p!=NULL; q=p,p=p->next_ )
						if( &p->obj_ == &*d )
							break;
					delete p;
					q->next_ = NULL;
				}
		}

private:

	Node* top_;
	NOCOPY(olist<T>);
};



//=========================================================================

}      // namespace ki
#endif // _KILIB_KTL_ARRAY_H_
